/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import com.bitctrl.Constants;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.LinguistischeVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.Term;
import de.bsvrz.iav.fuzzylib.fuzzylib.inferenz.Wissensbasis;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.FuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.OdFuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.PdFuzzyRegelbasis;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdmsFahrBahnZustandFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdmsGlaetteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdmsSichtVerhaeltnisseFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsHelligkeitFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsNiederschlagsIntensitaetFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsSichtWeiteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsWindGeschwindigkeitMittelWertFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsWindRichtungFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdAnteilLkw;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdBemessungsDichte;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdGeschwindkeit;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsStaerkeStunde;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsUnruhe;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsZustand;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsZustandUnruhe;
import de.bsvrz.iav.fuzzylib.interpreter.logik.FuzzyLogikHandler;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.bitctrl.interpreter.NichtWert;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.OnlineDatensatz;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.fachmodellglobal.zustaende.GueteVerfahren;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsHelligkeit;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsNiederschlagsIntensitaet;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsSichtWeite;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindGeschwindigkeitMittelWert;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindRichtung;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenMessStelle;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenSensor;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdStoerfallZustand;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.zustaende.StoerfallSituation;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Submodul Regelbasis. Empfängt Fuzzy-Wertewerte vom Datenverteiler, verknüpft
 * diese mit gegebenen Regeln und publiziert die neuen Fuzzy-Werte mit Hilfe des
 * Datenverteilers.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Regelbasis implements DatensatzUpdateListener {

	/** Logger. */
	private final Debug log = Debug.getLogger();

	/** Maximale Anzahl der gecachten Zyklen. */
	private final int maxZyklen;

	/** Lokaler Cache der Fuzzy-Werte. */
	private final Map<OdFuzzyWert, List<FuzzyVariable>> cache;

	/** Merkt sich den letzten Zeitstempel eines Störfallindikators. */
	private final Map<SystemObjekt, Long> letzterZeitstempel;

	/** Wird am Ende des Konstruktors auf {@code true} gesetzt. */
	private boolean bereit = false;

	/**
	 * Meldet sich als Empfänger der benötigten Fuzzy-Werte und als Quelle der
	 * daraus berechneten Fuzzy-Werte an.
	 *
	 * @param objekte
	 *            Zu bearbeitende Systemobjekte.
	 * @param maxZyklen
	 *            Maximale Anzahl von Zyklen, deren Fuzzy-Werte gecacht werden
	 *            sollen. Mit 1 wird nur der aktuelle Wert aufgehoben.
	 */
	Regelbasis(final Collection<SystemObjekt> objekte, final int maxZyklen) {
		int anzahlObjekte, anzahlFuzzyWerteEmpfaenger, anzahlFuzzyWerteQuelle,
				anzahlStoerfall;

		LogTools.log(log, SubSegAnaMsg.InfoStarteRegelbasis);

		this.maxZyklen = maxZyklen;
		cache = new HashMap<>();
		letzterZeitstempel = new HashMap<>();

		// Handler für fuzzy-logische Operatoren registrieren
		Operator.registerHandler(new FuzzyLogikHandler());

		// Als Quelle der Fuzzy-Werte anmelden
		anzahlObjekte = 0;
		anzahlFuzzyWerteEmpfaenger = 0;
		anzahlFuzzyWerteQuelle = 0;
		anzahlStoerfall = 0;
		for (final SystemObjekt so : objekte) {
			if (so instanceof MessQuerschnittAllgemein) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdAnteilLkw.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdBemessungsDichte.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdGeschwindkeit.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdVerkehrsStaerkeStunde.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdVerkehrsUnruhe.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdVerkehrsZustand.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdVerkehrsZustandUnruhe.class);
				anzahlStoerfall += anmeldenSender(so, OdStoerfallZustand.class);
			} else if (so instanceof StrassenTeilSegment) {
				final StrassenTeilSegment sts = (StrassenTeilSegment) so;
				if (sts.getMessQuerschnitte().isEmpty()) {
					// STS ohne MQ können ignoriert werden
					continue;
				}

				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdAnteilLkw.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdBemessungsDichte.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdGeschwindkeit.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdVerkehrsStaerkeStunde.class);
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdVerkehrsUnruhe.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdVerkehrsZustand.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdVerkehrsZustandUnruhe.class);
				anzahlStoerfall += anmeldenSender(so, OdStoerfallZustand.class);
			} else if (so instanceof UfdsHelligkeit) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdsHelligkeitFuzzy.class);
			} else if (so instanceof UfdsNiederschlagsIntensitaet) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdsNiederschlagsIntensitaetFuzzy.class);
			} else if (so instanceof UfdsWindGeschwindigkeitMittelWert) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdsWindGeschwindigkeitMittelWertFuzzy.class);
			} else if (so instanceof UfdsSichtWeite) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdsSichtWeiteFuzzy.class);
			} else if (so instanceof UfdsWindRichtung) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdsWindRichtungFuzzy.class);
			} else if (so instanceof UmfeldDatenMessStelle) {
				anzahlObjekte++;
				anzahlFuzzyWerteEmpfaenger += anmeldenEmpfaenger(so,
						OdUfdmsGlaetteFuzzy.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdUfdmsFahrBahnZustandFuzzy.class);
				anzahlFuzzyWerteQuelle += anmeldenSender(so,
						OdUfdmsSichtVerhaeltnisseFuzzy.class);
			} else {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.FinerIgnoriereUnbekanntesSystemObject, so);
			}
		}

		ObjektFactory.getInstanz().getVerbindung()
				.sleep(3 * Constants.MILLIS_PER_SECOND);
		synchronized (this) {
			bereit = true;
		}
		LogTools.log(log, SubSegAnaMsg.ConfigRegelbasisBereit, anzahlObjekte,
				anzahlFuzzyWerteEmpfaenger, anzahlFuzzyWerteQuelle,
				anzahlStoerfall);
	}

	@Override
	public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
		synchronized (this) {
			// Daten werden erst verarbeitet, wenn Konstruktor fertig ist.
			if (!bereit) {
				return;
			}
		}

		final SystemObjekt so = event.getObjekt();
		Map<String, FuzzyVariable> folgerungen;
		Kontext kontext = null;

		LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
				SubSegAnaMsg.FineEmpfangeFuzzyWert, so, event.getDatum());

		// Neue Daten cachen
		cacheWerte((OdFuzzyWert) event.getDatensatz(),
				(FuzzyWert) event.getDatum());

		if (so.hasParameterDatensatz(PdFuzzyRegelbasis.class)) {
			// I. d. R. hängt die Wissensbasis am Systemobjekt

			PdFuzzyRegelbasis.Daten datumWB;
			Wissensbasis wb;

			datumWB = so.getParameterDatensatz(PdFuzzyRegelbasis.class)
					.getDatum();
			if (!datumWB.isValid()) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.WarningUngueltigerWissensbasisParameter,
						so, event.getDatensatz().getAttributGruppe());
				return;
			}

			wb = datumWB.getWissensbasis();
			kontext = erstelleKontext(so);

			// Sonderfall Umfelddatenmessstelle
			if (so instanceof UmfeldDatenMessStelle) {
				UmfeldDatenMessStelle udms;

				// Alle angeschlossenen Umfelddatensensoren ergänzen
				udms = (UmfeldDatenMessStelle) so;
				for (final UmfeldDatenSensor uds : udms
						.getUmfelddatensensoren()) {
					kontext.addKontext(erstelleKontext(uds));
				}
			}

			// Wissensbasis anwenden und Ergebnis verschicken
			try {
				folgerungen = wb.interpret(kontext);
				sendeFolgerungen(so, folgerungen,
						event.getDatum().getZeitstempel());
			} catch (final NoSuchElementException ex) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.ErrorInWissensbasis, so, ex);
			} catch (final InterpreterException ex) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.ErrorInWissensbasis, so, ex);
			}

		} else if (so instanceof UmfeldDatenSensor) {

			// Spezialfall Umfelddatensensor

			List<UmfeldDatenMessStelle> listeUDMS;

			// Umfelddatenmessstellen des Umfelddatensensors ermitteln
			listeUDMS = ((UmfeldDatenSensor) so).getUmfelddatenMessstellen();
			for (final UmfeldDatenMessStelle udms : listeUDMS) {

				PdFuzzyRegelbasis.Daten datumWB;
				Wissensbasis wb;

				datumWB = udms.getParameterDatensatz(PdFuzzyRegelbasis.class)
						.getDatum();
				if (!datumWB.isValid()) {
					LogTools.log(log,
							new BetriebsmeldungDaten(so.getSystemObject()),
							SubSegAnaMsg.WarningUngueltigerWissensbasisParameter,
							so, event.getDatensatz().getAttributGruppe());
					return;
				}

				wb = datumWB.getWissensbasis();
				kontext = erstelleKontext(udms);

				// Kontexte der angeschlossenen Umfelddatensensoren sammeln
				for (final UmfeldDatenSensor uds : udms
						.getUmfelddatensensoren()) {
					kontext.addKontext(erstelleKontext(uds));
				}

				// Wissensbasis anwenden und Ergebnis verschicken
				try {
					folgerungen = wb.interpret(kontext);
					sendeFolgerungen(udms, folgerungen,
							event.getDatum().getZeitstempel());
				} catch (final NoSuchElementException ex) {
					LogTools.log(log,
							new BetriebsmeldungDaten(so.getSystemObject()),
							SubSegAnaMsg.ErrorInWissensbasis, so, ex);
				} catch (final InterpreterException ex) {
					LogTools.log(log,
							new BetriebsmeldungDaten(so.getSystemObject()),
							SubSegAnaMsg.ErrorInWissensbasis, so, ex);
				}
			}
		}
	}

	/**
	 * Meldet sich als Empfänger eines Datensatzes eines Systemobjekts an.
	 *
	 * @param so
	 *            ein Systemobjekt.
	 * @param datensatz
	 *            der zu empfangene Datensatz.
	 * @return 1 beim erfolgreichen Anmelden, sonst 0.
	 */
	private int anmeldenEmpfaenger(final SystemObjekt so,
			final Class<? extends OnlineDatensatz<?>> datensatz) {
		OnlineDatensatz<?> od;

		od = so.getOnlineDatensatz(datensatz);
		LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
				SubSegAnaMsg.FinerAnmeldenEmpfaengerFuzzyWert, so, od);
		od.addUpdateListener(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(), this);
		return 1;
	}

	/**
	 * Meldet sich als Quelle eines Datensatzes eines Systemobjekts an.
	 *
	 * @param so
	 *            ein Systemobjekt.
	 * @param datensatz
	 *            der zu sendende Datensatz.
	 * @return 1 beim erfolgreichen Anmelden, sonst 0.
	 */
	private int anmeldenSender(final SystemObjekt so,
			final Class<? extends OnlineDatensatz<?>> datensatz) {
		if (OdStoerfallZustand.class.isAssignableFrom(datensatz)) {
			OnlineDatensatz<?> od;

			od = so.getOnlineDatensatz(datensatz);
			LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
					SubSegAnaMsg.FinerAnmeldenQuelleStoerfall, so);
			od.setQuelle(OdStoerfallZustand.Aspekte.StoerfallVerfahrenFuzzy
					.getAspekt(), true);
			try {
				so.getOnlineDatensatz(datensatz).anmeldenSender(
						OdStoerfallZustand.Aspekte.StoerfallVerfahrenFuzzy
								.getAspekt());
				return 1;
			} catch (final AnmeldeException ex) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.ErrorBeimAnmeldenQuelleStoerfall, so, ex);
				return 0;
			}
		} else if (OdFuzzyWert.class.isAssignableFrom(datensatz)) {
			OnlineDatensatz<?> od;

			od = so.getOnlineDatensatz(datensatz);
			LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
					SubSegAnaMsg.FinerAnmeldenQuelleFuzzyWert, so,
					od.getAttributGruppe());
			od.setQuelle(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(), true);
			try {
				so.getOnlineDatensatz(datensatz)
						.anmeldenSender(OdFuzzyWert.Aspekte.Fuzzy.getAspekt());
				return 1;
			} catch (final AnmeldeException ex) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.ErrorBeimAnmeldenQuelleFuzzyWert, so,
						od.getAttributGruppe(), ex);
				return 0;
			}
		} else {
			throw new IllegalStateException();
		}

	}

	/**
	 * Sichert alle neuen Fuzzy-Werte des Events im Cache.
	 *
	 * @param datensatz
	 *            der ge&auml;nderte Datensatz.
	 * @param fw
	 *            der neue Wert im Datensatz.
	 */
	private void cacheWerte(final OdFuzzyWert datensatz, final FuzzyWert fw) {
		// Ggf. Datensatz im Cache anlegen
		if (!cache.containsKey(datensatz)) {
			cache.put(datensatz, new LinkedList<FuzzyVariable>());
		}

		// Neuen Wert sichern
		final List<FuzzyVariable> zyklen = cache.get(datensatz);
		zyklen.add(0, fw.getFuzzyWert());
		if (zyklen.size() > maxZyklen) {
			zyklen.remove(maxZyklen);
		}

	}

	/**
	 * Sammelt alle gecachten Variablen für ein Fuzzy-Objekt.
	 *
	 * @param so
	 *            Das betrachtete Fuzzy-Objekt
	 * @return Interpreter-Kontext mit allen Variablen des Objekts oder
	 *         {@code null}, wenn die linguistische Variable einer Konklusion
	 *         unbekannt ist
	 */
	private Kontext erstelleKontext(final SystemObjekt so) {
		final Kontext kontext = new Kontext();
		Set<String> konklusionen;

		// Alle Eingangsvariablen initial auf undefiniert setzen
		if (so.hasParameterDatensatz(PdFuzzyRegelbasis.class)) {
			final Wissensbasis wb = so
					.getParameterDatensatz(PdFuzzyRegelbasis.class).getDatum()
					.getWissensbasis();

			konklusionen = wb.getKonklusionen();
			for (final String s : wb.getPraemissen()) {
				kontext.set(s, NichtWert.NICHTWERT);
			}
		} else {
			konklusionen = new HashSet<>();
		}

		// Alle Werte des Objekts durchlaufen
		for (final OnlineDatensatz<?> od : so.getOnlineDatensatz()) {
			if (od instanceof OdFuzzyWert) {
				final OdFuzzyWert fw = (OdFuzzyWert) od;

				if (konklusionen.contains(fw.getName())) {
					// Konklusionen ohne Wert aufnehmen
					LinguistischeVariable lv;

					lv = so.getParameterDatensatz(fw.getLinguistischeVariable())
							.getDatum().getLinguistischeVariable();

					final FuzzyVariable fuzzyfiziere = lv.fuzzyfiziere();
					if (fuzzyfiziere == null) {
						final String msg = so
								+ ": Die Fuzzy-Eingangsvariable \""
								+ lv.getName()
								+ " kann nicht erzeugt werden, weil die linguistische Variable keine Fuzzy-Sets hat!";
						log.error(msg);
						throw new IllegalStateException(msg);
					}
					kontext.set(fw.getName(), fuzzyfiziere);
				} else {
					// Variablen aus der Prämisse aus dem Cache lesen
					List<FuzzyVariable> zyklen = cache.get(fw);

					if (zyklen == null) {
						zyklen = new ArrayList<>();
					}

					// Alle Zyklen der Fuzzy-Variablen in Kontext aufnehmen
					for (int i = 0; i < zyklen.size(); i++) {
						FuzzyVariable fv;

						fv = zyklen.get(i);
						if (fv != null) {
							if (i > 0) {
								kontext.set(fw.getName() + i, fv);
							} else {
								kontext.set(fw.getName(), fv);
							}
						} else {
							kontext.set(fw.getName(), NichtWert.NICHTWERT);
						}
					}
				}
			}
		}

		return kontext;
	}

	/**
	 * Sendet die Folgerungen aus der Wissensbasis.
	 *
	 * @param so
	 *            das Systemobjekt, dessen Wissensbasis abgearbeitet wurde.
	 * @param folgerungen
	 *            die Folgerungen.
	 * @param zeitstempel
	 *            der Zeitstempel mit dem gesendet werden soll.
	 */
	private void sendeFolgerungen(final SystemObjekt so,
			final Map<String, FuzzyVariable> folgerungen,
			final long zeitstempel) {
		for (final String wert : folgerungen.keySet()) {
			for (final OnlineDatensatz<?> od : so.getOnlineDatensatz()) {
				if (od instanceof OdFuzzyWert) {
					final OdFuzzyWert fw = (OdFuzzyWert) od;

					if (fw.getName().equals(wert)) {
						// Richtiger Fuzzy-Wert gefunden

						// Daten verpacken ...
						final FuzzyWert datum = fw.erzeugeDatum();
						datum.setFuzzyWert(folgerungen.get(wert));
						datum.setZeitstempel(zeitstempel);

						// ... und verschicken
						try {
							fw.sendeDaten(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
									datum);
							LogTools.log(log,
									new BetriebsmeldungDaten(
											so.getSystemObject()),
									SubSegAnaMsg.FineSendeFuzzyWert, so,
									fw.getAttributGruppe(), datum);
						} catch (final DatensendeException ex) {
							LogTools.log(log,
									new BetriebsmeldungDaten(
											so.getSystemObject()),
									SubSegAnaMsg.ErrorBeimSendenFuzzyWert, so,
									fw.getAttributGruppe(), datum, ex);
						}

						// Störfallindikator verschicken
						if (wert.equals("VerkehrsZustand")) { //$NON-NLS-1$
							sendeStoerfallIndikator(fw, datum);
						}
					}
				}
			}
		}
	}

	/**
	 * Sendet den Störfallzustand eines Objekts. Dabei wird der Fuzzy-Wert des
	 * Verkehrszustands in einen Störfall transferiert.
	 *
	 * @param fw
	 *            der Fuzzy-Wert der auf einen Störfallzustand abgebildet wird.
	 * @param datum
	 *            der aktuelle Wert der abgebildet wird.
	 */
	private void sendeStoerfallIndikator(final OdFuzzyWert fw,
			final FuzzyWert datum) {
		OdStoerfallZustand od;
		OdStoerfallZustand.Daten zustand;
		Term situation;
		SystemObjekt so;

		so = fw.getObjekt();
		od = so.getOnlineDatensatz(OdStoerfallZustand.class);
		zustand = od.erzeugeDatum();
		zustand.setZeitstempel(datum.getZeitstempel());

		// Den Term mit der höchsten Zugehörigkeit finden
		situation = null;
		if (datum.getFuzzyWert() != null) {
			for (final Term term : datum.getFuzzyWert().getTerme()) {
				if (situation == null) {
					situation = term;
				} else if ((situation.getZugehoerigkeit() == null)
						|| (term.getZugehoerigkeit() == null)) {
					situation = null;
					break;
				} else if (term.getZugehoerigkeit() > situation
						.getZugehoerigkeit()) {
					situation = term;
				}
			}
		}

		if ((situation != null) && (situation.getZugehoerigkeit() == 0)) {
			// Workaround für den Fall, dass alle Zugehörigkeiten 0 sind. In dem
			// Fall wird keine Aussage getroffen.
			situation = null;
		}

		// Differenz der Zeitstempel des aktuellen und letzten
		// Störfallindikators bestimmen
		if (letzterZeitstempel.containsKey(so)) {
			zustand.setT(datum.getZeitstempel() - letzterZeitstempel.get(so));
			letzterZeitstempel.put(so, datum.getZeitstempel());
		} else {
			// Noch kein letzter Zeitstempel gesichert
			zustand.setT(0);
		}

		// Horizont = 0 steht für die Verwendung der Analysewerte
		zustand.setHorizont(0);

		// Güte-Index = Zugehörigkeit des Terms mit der höchsten Zugehörigkeit
		if (situation != null) {
			zustand.setGueteIndex(situation.getZugehoerigkeit());
		} else {
			zustand.setGueteIndex(1);
		}

		// Güte-Verfahren ist immer 0 => Standardverfahren
		zustand.setGueteVerfahren(GueteVerfahren.Standard);

		if (situation == null) {
			// Verkehrszustand bzw. Situation konnte nicht abgeleitet werden
			zustand.setSituation(StoerfallSituation.KEINE_AUSSAGE);
		} else if ("Gestaut".equals(situation.getName())) { //$NON-NLS-1$
			zustand.setSituation(StoerfallSituation.STAU);
		} else if ("Zähfließend".equals(situation.getName())) { //$NON-NLS-1$
			zustand.setSituation(StoerfallSituation.ZAEHER_VERKEHR);
		} else if ("Dicht".equals(situation.getName())) { //$NON-NLS-1$
			zustand.setSituation(StoerfallSituation.DICHTER_VERKEHR);
		} else if ("Frei".equals(situation.getName())) { //$NON-NLS-1$
			zustand.setSituation(StoerfallSituation.FREIER_VERKEHR);
		} else {
			// Da nicht alle Zustände abgebildet werden ist dies der
			// Default-Wert
			zustand.setSituation(StoerfallSituation.KEINE_AUSSAGE);
		}

		try {
			od.sendeDaten(OdStoerfallZustand.Aspekte.StoerfallVerfahrenFuzzy
					.getAspekt(), zustand);
			LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
					SubSegAnaMsg.FineSendeStoerfall, so, zustand);
		} catch (final DatensendeException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
					SubSegAnaMsg.ErrorBeimSendenStoerfall, so, zustand, ex);
		}
	}

}
